@@ -1,7 +1,7 @@
 /*
  *  RODDC/src/RODBC.c 
  *         M. Lapsley Copyright (C) 1999-2002
- *         B. D. Ripley  Copyright (C) 2002-2016
+ *         B. D. Ripley  Copyright (C) 2002-2022
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -50,7 +50,8 @@ __declspec(dllimport) window RConsole;
 #include <sqlext.h>
 
 #include <R.h>
-#include <Rdefines.h>
+#include <Rinternals.h>
+//#include <Rdefines.h>
 
 #ifdef ENABLE_NLS
 #include <libintl.h>
@@ -225,7 +226,7 @@ SEXP RODBCDriverConnect(SEXP connection, SEXP id, SEXP useNRows, SEXP ReadOnly)
 	UNPROTECT(1);
 	return ans;
     }
-    thisHandle = Calloc(1, RODBCHandle);
+    thisHandle = R_Calloc(1, RODBCHandle);
     ++nChannels;
 
     odbcInit();
@@ -333,9 +334,9 @@ SEXP RODBCQuery(SEXP chan, SEXP query, SEXP sRows)
 			    (SQLCHAR *) translateChar(STRING_ELT(query, 0)),
 			    SQL_NTS);
 	if( res != SQL_SUCCESS && res != SQL_SUCCESS_WITH_INFO ) {
-	    char *message = Calloc(strlen(cquery)+50, char);
-	    sprintf(message,
-		    "[RODBC] ERROR: Could not SQLExecDirect '%s'", cquery);
+	    char *message = R_Calloc(strlen(cquery)+50, char);
+	    snprintf(message, strlen(cquery)+50,
+		     "[RODBC] ERROR: Could not SQLExecDirect '%s'", cquery);
 	    geterr(thisHandle);
 	    errlistAppend(thisHandle, message);
 	    (void)SQLFreeHandle(SQL_HANDLE_STMT, thisHandle->hStmt);
@@ -615,8 +616,8 @@ static void cachenbind_free(pRODBCHandle thisHandle)
     if(thisHandle->ColData) {
 	for (i = 0; i < thisHandle->nAllocated; i++)
 	    if(thisHandle->ColData[i].pData)
-		Free(thisHandle->ColData[i].pData);
-	Free(thisHandle->ColData);
+		R_Free(thisHandle->ColData[i].pData);
+	R_Free(thisHandle->ColData);
 	thisHandle->ColData = NULL; /* to be sure */
     }   
 }
@@ -657,7 +658,7 @@ static int cachenbind(pRODBCHandle thisHandle, int nRows)
     /* Allocate storage for ColData array,
        first freeing what was there before */
     cachenbind_free(thisHandle);
-    thisHandle->ColData = Calloc(NCOLS, COLUMNS);
+    thisHandle->ColData = R_Calloc(NCOLS, COLUMNS);
     /* this allocates Data as zero */
     thisHandle->nAllocated = NCOLS;
 
@@ -730,7 +731,7 @@ static int cachenbind(pRODBCHandle thisHandle, int nRows)
 	    SQLLEN datalen = thisHandle->ColData[i].ColSize;
 	    thisHandle->ColData[i].datalen = datalen;
 	    thisHandle->ColData[i].pData =
-		Calloc(nRows * (datalen + 1), char);
+		R_Calloc(nRows * (datalen + 1), char);
 	    BIND(SQL_C_BINARY, pData, datalen);
 	}
 	default:
@@ -740,7 +741,7 @@ static int cachenbind(pRODBCHandle thisHandle, int nRows)
 	    /* sanity check as the reports are sometimes unreliable */
 	    if (datalen > 65535) datalen = 65535;
 	    thisHandle->ColData[i].pData =
-		Calloc(nRows * (datalen + 1), char);
+		R_Calloc(nRows * (datalen + 1), char);
 	    thisHandle->ColData[i].datalen = datalen;
 	    BIND(SQL_C_CHAR, pData, datalen);
 	}
@@ -817,7 +818,7 @@ SEXP RODBCFetchRows(SEXP chan, SEXP max, SEXP bs, SEXP nas, SEXP believeNRows)
 		blksize = (buffsize < 100) ? 100: buffsize;
 	    }
 	} else {
-	    if(!maximum || maximum > NROWS) maximum = NROWS;
+	    if(!maximum || maximum > NROWS) maximum = (int) NROWS;
 	    blksize = maximum;
 	}
 	for(i = 0; i < nc; i++) {
@@ -916,7 +917,8 @@ SEXP RODBCFetchRows(SEXP chan, SEXP max, SEXP bs, SEXP nas, SEXP believeNRows)
 			SEXP this = (len == SQL_NULL_DATA) ?
 			    allocVector(RAWSXP, 0) :
 			    mkRaw(thisHandle->ColData[i].pData + 
-				  (thisHandle->ColData[i].datalen * row), len);
+				  (thisHandle->ColData[i].datalen * row),
+				  (unsigned int) len);
 			SET_VECTOR_ELT(VECTOR_ELT(data, i), j-1, this);
 			break;
 		    }
@@ -952,7 +954,8 @@ SEXP RODBCFetchRows(SEXP chan, SEXP max, SEXP bs, SEXP nas, SEXP believeNRows)
     PROTECT(names = allocVector(STRSXP, 2));
     SET_STRING_ELT(names, 0, mkChar("data"));
     SET_STRING_ELT(names, 1, mkChar("stat"));
-    SET_NAMES(ans, names);
+    setAttrib(ans, R_NamesSymbol, names);
+//    SET_NAMES(ans, names);
     UNPROTECT(3); /* ans data names */
     return ans;
 }
@@ -1022,14 +1025,14 @@ SEXP RODBCUpdate(SEXP chan, SEXP query, SEXP data, SEXP dataseq,
 		 SEXP params, SEXP test)
 {
     pRODBCHandle thisHandle = R_ExternalPtrAddr(chan);
-    int rows, i, j, k, stat = 1, vtest = asInteger(test), nparams;
+    int rows, i, k, stat = 1, vtest = asInteger(test), nparams;
     int *sequence = INTEGER(dataseq);
     const char *cquery = translateChar(STRING_ELT(query, 0));
     SQLRETURN res = 0; /* -Wall */
     SEXP paramnames = VECTOR_ELT(params, 0);
 
     nparams = length(paramnames);
-    NCOLS = nparams;
+    NCOLS = (SQLSMALLINT)nparams;
 
     clearresults(thisHandle);
 
@@ -1042,9 +1045,9 @@ SEXP RODBCUpdate(SEXP chan, SEXP query, SEXP data, SEXP dataseq,
     res = SQLPrepare( thisHandle->hStmt, (SQLCHAR *) cquery,
 		      strlen(cquery) );
     if( res != SQL_SUCCESS && res != SQL_SUCCESS_WITH_INFO ) {
-	char *message = Calloc(strlen(cquery)+50, char);
-	sprintf(message,
-		"[RODBC] ERROR: Could not SQLPrepare '%s'", cquery);
+	char *message = R_Calloc(strlen(cquery)+50, char);
+	snprintf(message,strlen(cquery)+50,
+		 "[RODBC] ERROR: Could not SQLPrepare '%s'", cquery);
 	geterr(thisHandle);
 	errlistAppend(thisHandle, message);
 	(void)SQLFreeHandle(SQL_HANDLE_STMT, thisHandle->hStmt);
@@ -1055,23 +1058,22 @@ SEXP RODBCUpdate(SEXP chan, SEXP query, SEXP data, SEXP dataseq,
     /* Allocate storage for ColData array,
        first freeing what was there last */
     cachenbind_free(thisHandle);
-    thisHandle->ColData = Calloc(NCOLS, COLUMNS);
+    thisHandle->ColData = R_Calloc(NCOLS, COLUMNS);
     /* this allocates Data as zero */
     thisHandle->nAllocated = NCOLS;
 
     /* extract the column data  */
-    for(j = 0; j < nparams; j++) {
-	int tmp;
+    for(short j = 0; j < nparams; j++) {
 	strcpy((char *) thisHandle->ColData[j].ColName,
 	       translateChar(STRING_ELT(paramnames, j)));
 	thisHandle->ColData[j].DataType = INTEGER(VECTOR_ELT(params, 1))[j];
 	thisHandle->ColData[j].ColSize = INTEGER(VECTOR_ELT(params, 2))[j];
 	/* I don't think this would be NA, but the code was here */
-	tmp  = INTEGER(VECTOR_ELT(params, 3))[j];
+	int tmp  = INTEGER(VECTOR_ELT(params, 3))[j];
 	thisHandle->ColData[j].DecimalDigits = (tmp == NA_INTEGER) ? 0 : tmp;
 
 	if(vtest)
-	    Rprintf("Binding: '%s' DataType %d, ColSize %d\n",
+	    Rprintf("Binding: '%s' DataType %d, ColSize %ul\n",
 		    (char *) thisHandle->ColData[j].ColName,
 		    thisHandle->ColData[j].DataType,
 		    thisHandle->ColData[j].ColSize);
@@ -1100,10 +1102,10 @@ SEXP RODBCUpdate(SEXP chan, SEXP query, SEXP data, SEXP dataseq,
 	    break;
 	default:
 	{
-	    int datalen = thisHandle->ColData[j].ColSize;
+	    unsigned long datalen = thisHandle->ColData[j].ColSize;
 	    /* Why change here but not when sending the data? */
 	    if (datalen <= 0) datalen = 1024;
-	    thisHandle->ColData[j].pData = Calloc(datalen+1, char);
+	    thisHandle->ColData[j].pData = R_Calloc(datalen+1, char);
 	    res = SQLBindParameter(thisHandle->hStmt,
 				   j+1, SQL_PARAM_INPUT, SQL_C_CHAR,
 				   thisHandle->ColData[j].DataType,
@@ -1128,7 +1130,7 @@ SEXP RODBCUpdate(SEXP chan, SEXP query, SEXP data, SEXP dataseq,
     if(vtest) Rprintf("Parameters:\n");
     rows = LENGTH(VECTOR_ELT(data, 0));
     for(i = 0; i < rows; i++) {
-	for(j = 0; j < LENGTH(data); j++) {
+	for(int j = 0; j < LENGTH(data); j++) {
 	    k = sequence[j]; /* get the right column */
 	    switch(TYPEOF(VECTOR_ELT(data, k))) { 
 	    case REALSXP:
@@ -1158,7 +1160,7 @@ SEXP RODBCUpdate(SEXP chan, SEXP query, SEXP data, SEXP dataseq,
 	    default:
 	    {
 		const char *cData = translateChar(STRING_ELT(VECTOR_ELT(data, k), i));
-		int datalen = thisHandle->ColData[j].ColSize;
+		unsigned long datalen = thisHandle->ColData[j].ColSize;
 		strncpy(thisHandle->ColData[j].pData, cData, datalen);
 		thisHandle->ColData[j].pData[datalen] = '\0';
 		if(strlen(cData) > datalen)
@@ -1222,7 +1224,7 @@ static int inRODBCClose(pRODBCHandle thisHandle)
     cachenbind_free(thisHandle);
     errorFree(thisHandle->msglist);
     R_ClearExternalPtr(thisHandle->extPtr);
-    Free(thisHandle);
+    R_Free(thisHandle);
     return success;
 }
 
@@ -1278,7 +1280,8 @@ geterr(pRODBCHandle thisHandle)
 				&MsgLen);
 
 	if(retval != SQL_SUCCESS && retval != SQL_SUCCESS_WITH_INFO) break;
-	sprintf(message,"%s %d %s", sqlstate, (int)NativeError, msg);
+	snprintf(message, SQL_MAX_MESSAGE_LENGTH+16,
+		 "%s %d %s", sqlstate, (int)NativeError, msg);
 	errlistAppend(thisHandle, message);
 	// i++;
     }
@@ -1292,7 +1295,7 @@ geterr(pRODBCHandle thisHandle)
 static char *mystrdup(const char *s)
 {
     char *s2;
-    s2 = Calloc(strlen(s) + 1, char);
+    s2 = R_Calloc(strlen(s) + 1, char);
     strcpy(s2, s);
     return s2;
 }
@@ -1315,10 +1318,10 @@ static void errlistAppend(pRODBCHandle thisHandle, const char *string)
 	    if(root->next) root = root->next;
 	    else break;
 	}
-	root->next = Calloc(1, SQLMSG);
+	root->next = R_Calloc(1, SQLMSG);
 	root = root->next;
     } else {
-	root = thisHandle->msglist = Calloc(1, SQLMSG);
+	root = thisHandle->msglist = R_Calloc(1, SQLMSG);
     }
     root->next = NULL;
     root->message = buffer;
@@ -1401,8 +1404,8 @@ static void errorFree(SQLMSG *node)
     if(node->next)
 	errorFree(node->next);
     if(node) {
-	Free(node->message);
-	Free(node);
+	R_Free(node->message);
+	R_Free(node);
 	node = NULL;
     }
 }
@@ -1484,7 +1487,8 @@ SEXP RODBCListDataSources(SEXP stype)
 				sizeof(szDescription), NULL);
 	if(retval == SQL_NO_DATA) break;
 	if(retval != SQL_SUCCESS && retval != SQL_SUCCESS_WITH_INFO) {
-	    sprintf(message, "SQLDataSources returned: %d", retval);
+	    snprintf(message, SQL_MAX_DSN_LENGTH+101,
+		     "SQLDataSources returned: %d", retval);
 	    SET_STRING_ELT(ans, i, mkChar(message));
 	} else {
 	    SET_STRING_ELT(nm, i, mkChar((char *)szDSN));
@@ -1499,8 +1503,8 @@ SEXP RODBCListDataSources(SEXP stype)
 	}
     } while(retval == SQL_SUCCESS || retval == SQL_SUCCESS_WITH_INFO);
 
-    ans = lengthgets(ans, i);
-    nm = lengthgets(nm, i);
+    REPROTECT(ans = lengthgets(ans, i), pidx);
+    REPROTECT(nm = lengthgets(nm, i), nidx);
     setAttrib(ans, R_NamesSymbol, nm);
     UNPROTECT(2);
     return ans;
